@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # This file is part of beets.
 # Copyright 2016, Adrian Sampson.
 #
@@ -17,17 +16,15 @@
 music and items' embedded album art.
 """
 
-from __future__ import division, absolute_import, print_function
 
 import subprocess
 import platform
 from tempfile import NamedTemporaryFile
-import imghdr
 import os
 
 from beets.util import displayable_path, syspath, bytestring_path
 from beets.util.artresizer import ArtResizer
-from beets import mediafile
+import mediafile
 
 
 def mediafile_image(image_path, maxwidth=None):
@@ -44,7 +41,7 @@ def get_art(log, item):
     try:
         mf = mediafile.MediaFile(syspath(item.path))
     except mediafile.UnreadableFileError as exc:
-        log.warning(u'Could not extract art from {0}: {1}',
+        log.warning('Could not extract art from {0}: {1}',
                     displayable_path(item.path), exc)
         return
 
@@ -52,26 +49,27 @@ def get_art(log, item):
 
 
 def embed_item(log, item, imagepath, maxwidth=None, itempath=None,
-               compare_threshold=0, ifempty=False, as_album=False):
+               compare_threshold=0, ifempty=False, as_album=False, id3v23=None,
+               quality=0):
     """Embed an image into the item's media file.
     """
     # Conditions and filters.
     if compare_threshold:
         if not check_art_similarity(log, item, imagepath, compare_threshold):
-            log.info(u'Image not similar; skipping.')
+            log.info('Image not similar; skipping.')
             return
     if ifempty and get_art(log, item):
-            log.info(u'media file already contained art')
-            return
+        log.info('media file already contained art')
+        return
     if maxwidth and not as_album:
-        imagepath = resize_image(log, imagepath, maxwidth)
+        imagepath = resize_image(log, imagepath, maxwidth, quality)
 
     # Get the `Image` object from the file.
     try:
-        log.debug(u'embedding {0}', displayable_path(imagepath))
+        log.debug('embedding {0}', displayable_path(imagepath))
         image = mediafile_image(imagepath, maxwidth)
-    except IOError as exc:
-        log.warning(u'could not read image file: {0}', exc)
+    except OSError as exc:
+        log.warning('could not read image file: {0}', exc)
         return
 
     # Make sure the image kind is safe (some formats only support PNG
@@ -81,36 +79,39 @@ def embed_item(log, item, imagepath, maxwidth=None, itempath=None,
                  image.mime_type)
         return
 
-    item.try_write(path=itempath, tags={'images': [image]})
+    item.try_write(path=itempath, tags={'images': [image]}, id3v23=id3v23)
 
 
-def embed_album(log, album, maxwidth=None, quiet=False,
-                compare_threshold=0, ifempty=False):
+def embed_album(log, album, maxwidth=None, quiet=False, compare_threshold=0,
+                ifempty=False, quality=0):
     """Embed album art into all of the album's items.
     """
     imagepath = album.artpath
     if not imagepath:
-        log.info(u'No album art present for {0}', album)
+        log.info('No album art present for {0}', album)
         return
     if not os.path.isfile(syspath(imagepath)):
-        log.info(u'Album art not found at {0} for {1}',
+        log.info('Album art not found at {0} for {1}',
                  displayable_path(imagepath), album)
         return
     if maxwidth:
-        imagepath = resize_image(log, imagepath, maxwidth)
+        imagepath = resize_image(log, imagepath, maxwidth, quality)
 
-    log.info(u'Embedding album art into {0}', album)
+    log.info('Embedding album art into {0}', album)
 
     for item in album.items():
-        embed_item(log, item, imagepath, maxwidth, None,
-                   compare_threshold, ifempty, as_album=True)
+        embed_item(log, item, imagepath, maxwidth, None, compare_threshold,
+                   ifempty, as_album=True, quality=quality)
 
 
-def resize_image(log, imagepath, maxwidth):
-    """Returns path to an image resized to maxwidth.
+def resize_image(log, imagepath, maxwidth, quality):
+    """Returns path to an image resized to maxwidth and encoded with the
+    specified quality level.
     """
-    log.debug(u'Resizing album art to {0} pixels wide', maxwidth)
-    imagepath = ArtResizer.shared.resize(maxwidth, syspath(imagepath))
+    log.debug('Resizing album art to {0} pixels wide and encoding at quality \
+              level {1}', maxwidth, quality)
+    imagepath = ArtResizer.shared.resize(maxwidth, syspath(imagepath),
+                                         quality=quality)
     return imagepath
 
 
@@ -128,11 +129,11 @@ def check_art_similarity(log, item, imagepath, compare_threshold):
             # to grayscale and then pipe them into the `compare` command.
             # On Windows, ImageMagick doesn't support the magic \\?\ prefix
             # on paths, so we pass `prefix=False` to `syspath`.
-            convert_cmd = [b'convert', syspath(imagepath, prefix=False),
+            convert_cmd = ['convert', syspath(imagepath, prefix=False),
                            syspath(art, prefix=False),
-                           b'-colorspace', b'gray', b'MIFF:-']
-            compare_cmd = [b'compare', b'-metric', b'PHASH', b'-', b'null:']
-            log.debug(u'comparing images with pipeline {} | {}',
+                           '-colorspace', 'gray', 'MIFF:-']
+            compare_cmd = ['compare', '-metric', 'PHASH', '-', 'null:']
+            log.debug('comparing images with pipeline {} | {}',
                       convert_cmd, compare_cmd)
             convert_proc = subprocess.Popen(
                 convert_cmd,
@@ -156,7 +157,7 @@ def check_art_similarity(log, item, imagepath, compare_threshold):
             convert_proc.wait()
             if convert_proc.returncode:
                 log.debug(
-                    u'ImageMagick convert failed with status {}: {!r}',
+                    'ImageMagick convert failed with status {}: {!r}',
                     convert_proc.returncode,
                     convert_stderr,
                 )
@@ -166,7 +167,7 @@ def check_art_similarity(log, item, imagepath, compare_threshold):
             stdout, stderr = compare_proc.communicate()
             if compare_proc.returncode:
                 if compare_proc.returncode != 1:
-                    log.debug(u'ImageMagick compare failed: {0}, {1}',
+                    log.debug('ImageMagick compare failed: {0}, {1}',
                               displayable_path(imagepath),
                               displayable_path(art))
                     return
@@ -177,10 +178,10 @@ def check_art_similarity(log, item, imagepath, compare_threshold):
             try:
                 phash_diff = float(out_str)
             except ValueError:
-                log.debug(u'IM output is not a number: {0!r}', out_str)
+                log.debug('IM output is not a number: {0!r}', out_str)
                 return
 
-            log.debug(u'ImageMagick copmare score: {0}', phash_diff)
+            log.debug('ImageMagick compare score: {0}', phash_diff)
             return phash_diff <= compare_threshold
 
     return True
@@ -190,18 +191,18 @@ def extract(log, outpath, item):
     art = get_art(log, item)
     outpath = bytestring_path(outpath)
     if not art:
-        log.info(u'No album art present in {0}, skipping.', item)
+        log.info('No album art present in {0}, skipping.', item)
         return
 
     # Add an extension to the filename.
-    ext = imghdr.what(None, h=art)
+    ext = mediafile.image_extension(art)
     if not ext:
-        log.warning(u'Unknown image type in {0}.',
+        log.warning('Unknown image type in {0}.',
                     displayable_path(item.path))
         return
     outpath += bytestring_path('.' + ext)
 
-    log.info(u'Extracting album art from: {0} to: {1}',
+    log.info('Extracting album art from: {0} to: {1}',
              item, displayable_path(outpath))
     with open(syspath(outpath), 'wb') as f:
         f.write(art)
@@ -217,7 +218,7 @@ def extract_first(log, outpath, items):
 
 def clear(log, lib, query):
     items = lib.items(query)
-    log.info(u'Clearing album art from {0} items', len(items))
+    log.info('Clearing album art from {0} items', len(items))
     for item in items:
-        log.debug(u'Clearing art for {0}', item)
+        log.debug('Clearing art for {0}', item)
         item.try_write(tags={'images': None})
