@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # This file is part of beets.
 # Copyright 2016, Adrian Sampson.
 #
@@ -16,7 +15,6 @@
 """A simple utility for constructing filesystem-like trees from beets
 libraries.
 """
-from __future__ import division, absolute_import, print_function
 
 from collections import namedtuple
 from beets import util
