@@ -1,5 +1,5 @@
 # This file is part of beets.
-# Copyright 2014, Fabrice Laporte.
+# Copyright 2016, Fabrice Laporte.
 #
 # Permission is hereby granted, free of charge, to any person obtaining
 # a copy of this software and associated documentation files (the
@@ -15,14 +15,16 @@
 """Provides the %bucket{} function for path formatting.
 """
 
+
 from datetime import datetime
-import logging
 import re
 import string
-from itertools import tee, izip
+from itertools import tee
+
 from beets import plugins, ui
 
-log = logging.getLogger('beets')
+
+ASCII_DIGITS = string.digits + string.ascii_lowercase
 
 
 class BucketError(Exception):
@@ -33,7 +35,7 @@ def pairwise(iterable):
     "s -> (s0,s1), (s1,s2), (s2, s3), ..."
     a, b = tee(iterable)
     next(b, None)
-    return izip(a, b)
+    return zip(a, b)
 
 
 def span_from_str(span_str):
@@ -55,7 +57,7 @@ def span_from_str(span_str):
                 d = (yearfrom - yearfrom % 100) + d
         return d
 
-    years = [int(x) for x in re.findall('\d+', span_str)]
+    years = [int(x) for x in re.findall(r'\d+', span_str)]
     if not years:
         raise ui.UserError("invalid range defined for year bucket '%s': no "
                            "year found" % span_str)
@@ -117,16 +119,16 @@ def build_year_spans(year_spans_str):
 def str2fmt(s):
     """Deduces formatting syntax from a span string.
     """
-    regex = re.compile("(?P<bef>\D*)(?P<fromyear>\d+)(?P<sep>\D*)"
-                       "(?P<toyear>\d*)(?P<after>\D*)")
+    regex = re.compile(r"(?P<bef>\D*)(?P<fromyear>\d+)(?P<sep>\D*)"
+                       r"(?P<toyear>\d*)(?P<after>\D*)")
     m = re.match(regex, s)
 
     res = {'fromnchars': len(m.group('fromyear')),
            'tonchars': len(m.group('toyear'))}
-    res['fmt'] = "%s%%s%s%s%s" % (m.group('bef'),
-                                  m.group('sep'),
-                                  '%s' if res['tonchars'] else '',
-                                  m.group('after'))
+    res['fmt'] = "{}%s{}{}{}".format(m.group('bef'),
+                                     m.group('sep'),
+                                     '%s' if res['tonchars'] else '',
+                                     m.group('after'))
     return res
 
 
@@ -136,6 +138,7 @@ def format_span(fmt, yearfrom, yearto, fromnchars, tonchars):
     args = (str(yearfrom)[-fromnchars:])
     if tonchars:
         args = (str(yearfrom)[-fromnchars:], str(yearto)[-tonchars:])
+
     return fmt % args
 
 
@@ -154,23 +157,23 @@ def build_alpha_spans(alpha_spans_str, alpha_regexs):
     [from...to]
     """
     spans = []
-    ASCII_DIGITS = string.digits + string.ascii_lowercase
+
     for elem in alpha_spans_str:
         if elem in alpha_regexs:
             spans.append(re.compile(alpha_regexs[elem]))
         else:
             bucket = sorted([x for x in elem.lower() if x.isalnum()])
             if bucket:
-                beginIdx = ASCII_DIGITS.index(bucket[0])
-                endIdx = ASCII_DIGITS.index(bucket[-1])
+                begin_index = ASCII_DIGITS.index(bucket[0])
+                end_index = ASCII_DIGITS.index(bucket[-1])
             else:
                 raise ui.UserError("invalid range defined for alpha bucket "
                                    "'%s': no alphanumeric character found" %
                                    elem)
             spans.append(
                 re.compile(
-                    "^[" + ASCII_DIGITS[beginIdx:endIdx + 1] +
-                    ASCII_DIGITS[beginIdx:endIdx + 1].upper() + "]"
+                    "^[" + ASCII_DIGITS[begin_index:end_index + 1] +
+                    ASCII_DIGITS[begin_index:end_index + 1].upper() + "]"
                 )
             )
     return spans
@@ -178,7 +181,7 @@ def build_alpha_spans(alpha_spans_str, alpha_regexs):
 
 class BucketPlugin(plugins.BeetsPlugin):
     def __init__(self):
-        super(BucketPlugin, self).__init__()
+        super().__init__()
         self.template_funcs['bucket'] = self._tmpl_bucket
 
         self.config.add({
