@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
 # This file is part of beets.
-# Copyright 2015, Bruno Cauet
+# Copyright 2016, Bruno Cauet
 #
 # Permission is hereby granted, free of charge, to any person obtaining
 # a copy of this software and associated documentation files (the
@@ -19,8 +18,6 @@ This plugin is POSIX-only.
 Spec: standards.freedesktop.org/thumbnail-spec/latest/index.html
 """
 
-from __future__ import (division, absolute_import, print_function,
-                        unicode_literals)
 
 from hashlib import md5
 import os
@@ -35,7 +32,7 @@ from xdg import BaseDirectory
 from beets.plugins import BeetsPlugin
 from beets.ui import Subcommand, decargs
 from beets import util
-from beets.util.artresizer import ArtResizer, has_IM, has_PIL
+from beets.util.artresizer import ArtResizer, get_im_version, get_pil_version
 
 
 BASE_DIR = os.path.join(BaseDirectory.xdg_cache_home, "thumbnails")
@@ -45,7 +42,7 @@ LARGE_DIR = util.bytestring_path(os.path.join(BASE_DIR, "large"))
 
 class ThumbnailsPlugin(BeetsPlugin):
     def __init__(self):
-        super(ThumbnailsPlugin, self).__init__()
+        super().__init__()
         self.config.add({
             'auto': True,
             'force': False,
@@ -60,7 +57,8 @@ class ThumbnailsPlugin(BeetsPlugin):
         thumbnails_command = Subcommand("thumbnails",
                                         help="Create album thumbnails")
         thumbnails_command.parser.add_option(
-            '-f', '--force', dest='force', action='store_true', default=False,
+            '-f', '--force',
+            dest='force', action='store_true', default=False,
             help='force regeneration of thumbnails deemed fine (existing & '
                  'recent enough)')
         thumbnails_command.parser.add_option(
@@ -92,11 +90,11 @@ class ThumbnailsPlugin(BeetsPlugin):
             if not os.path.exists(dir):
                 os.makedirs(dir)
 
-        if has_IM():
+        if get_im_version():
             self.write_metadata = write_metadata_im
             tool = "IM"
         else:
-            assert has_PIL()  # since we're local
+            assert get_pil_version()  # since we're local
             self.write_metadata = write_metadata_pil
             tool = "PIL"
         self._log.debug("using {0} to write metadata", tool)
@@ -112,9 +110,9 @@ class ThumbnailsPlugin(BeetsPlugin):
     def process_album(self, album):
         """Produce thumbnails for the album folder.
         """
-        self._log.debug(u'generating thumbnail for {0}', album)
+        self._log.debug('generating thumbnail for {0}', album)
         if not album.artpath:
-            self._log.info(u'album {0} has no art', album)
+            self._log.info('album {0} has no art', album)
             return
 
         if self.config['dolphin']:
@@ -159,18 +157,19 @@ class ThumbnailsPlugin(BeetsPlugin):
 
     def thumbnail_file_name(self, path):
         """Compute the thumbnail file name
-        See http://standards.freedesktop.org/thumbnail-spec/latest/x227.html
+        See https://standards.freedesktop.org/thumbnail-spec/latest/x227.html
         """
         uri = self.get_uri(path)
-        hash = md5(uri).hexdigest()
-        return b"{0}.png".format(hash)
+        hash = md5(uri.encode('utf-8')).hexdigest()
+        return util.bytestring_path(f"{hash}.png")
 
     def add_tags(self, album, image_path):
         """Write required metadata to the thumbnail
-        See http://standards.freedesktop.org/thumbnail-spec/latest/x142.html
+        See https://standards.freedesktop.org/thumbnail-spec/latest/x142.html
         """
+        mtime = os.stat(album.artpath).st_mtime
         metadata = {"Thumb::URI": self.get_uri(album.artpath),
-                    "Thumb::MTime": unicode(os.stat(album.artpath).st_mtime)}
+                    "Thumb::MTime": str(mtime)}
         try:
             self.write_metadata(image_path, metadata)
         except Exception:
@@ -183,7 +182,8 @@ class ThumbnailsPlugin(BeetsPlugin):
             return
         artfile = os.path.split(album.artpath)[1]
         with open(outfilename, 'w') as f:
-            f.write(b"[Desktop Entry]\nIcon=./{0}".format(artfile))
+            f.write('[Desktop Entry]\n')
+            f.write('Icon=./{}'.format(artfile.decode('utf-8')))
             f.close()
         self._log.debug("Wrote file {0}", util.displayable_path(outfilename))
 
@@ -208,7 +208,7 @@ def write_metadata_pil(file, metadata):
     return True
 
 
-class URIGetter(object):
+class URIGetter:
     available = False
     name = "Abstract base"
 
@@ -221,7 +221,7 @@ class PathlibURI(URIGetter):
     name = "Python Pathlib"
 
     def uri(self, path):
-        return PurePosixPath(path).as_uri()
+        return PurePosixPath(util.py3_path(path)).as_uri()
 
 
 def copy_c_string(c_string):
@@ -232,7 +232,7 @@ def copy_c_string(c_string):
     # work. A more surefire way would be to allocate a ctypes buffer and copy
     # the data with `memcpy` or somesuch.
     s = ctypes.cast(c_string, ctypes.c_char_p).value
-    return '' + s
+    return b'' + s
 
 
 class GioURI(URIGetter):
@@ -266,24 +266,26 @@ class GioURI(URIGetter):
     def uri(self, path):
         g_file_ptr = self.libgio.g_file_new_for_path(path)
         if not g_file_ptr:
-            raise RuntimeError("No gfile pointer received for {0}".format(
+            raise RuntimeError("No gfile pointer received for {}".format(
                 util.displayable_path(path)))
 
         try:
             uri_ptr = self.libgio.g_file_get_uri(g_file_ptr)
-        except:
-            raise
         finally:
             self.libgio.g_object_unref(g_file_ptr)
         if not uri_ptr:
             self.libgio.g_free(uri_ptr)
             raise RuntimeError("No URI received from the gfile pointer for "
-                               "{0}".format(util.displayable_path(path)))
+                               "{}".format(util.displayable_path(path)))
 
         try:
             uri = copy_c_string(uri_ptr)
-        except:
-            raise
         finally:
             self.libgio.g_free(uri_ptr)
-        return uri
+
+        try:
+            return uri.decode(util._fsencoding())
+        except UnicodeDecodeError:
+            raise RuntimeError(
+                f"Could not decode filename from GIO: {uri!r}"
+            )
