@@ -1,5 +1,5 @@
 # This file is part of beets.
-# Copyright 2015, Thomas Scholtes.
+# Copyright 2016, Thomas Scholtes.
 #
 # Permission is hereby granted, free of charge, to any person obtaining
 # a copy of this software and associated documentation files (the
@@ -15,25 +15,23 @@
 """Uses the `KeyFinder` program to add the `initial_key` field.
 """
 
-from __future__ import (division, absolute_import, print_function,
-                        unicode_literals)
 
+import os.path
 import subprocess
 
 from beets import ui
 from beets import util
 from beets.plugins import BeetsPlugin
-from beets import config
 
 
 class KeyFinderPlugin(BeetsPlugin):
 
     def __init__(self):
-        super(KeyFinderPlugin, self).__init__()
+        super().__init__()
         self.config.add({
-            u'bin': u'KeyFinder',
-            u'auto': True,
-            u'overwrite': False,
+            'bin': 'KeyFinder',
+            'auto': True,
+            'overwrite': False,
         })
 
         if self.config['auto'].get(bool):
@@ -46,35 +44,52 @@ class KeyFinderPlugin(BeetsPlugin):
         return [cmd]
 
     def command(self, lib, opts, args):
-        self.find_key(lib.items(ui.decargs(args)),
-                      write=config['import']['write'].get(bool))
+        self.find_key(lib.items(ui.decargs(args)), write=ui.should_write())
 
     def imported(self, session, task):
-        self.find_key(task.items)
+        self.find_key(task.imported_items())
 
     def find_key(self, items, write=False):
         overwrite = self.config['overwrite'].get(bool)
-        bin = util.bytestring_path(self.config['bin'].get(unicode))
+        command = [self.config['bin'].as_str()]
+        # The KeyFinder GUI program needs the -f flag before the path.
+        # keyfinder-cli is similar, but just wants the path with no flag.
+        if 'keyfinder-cli' not in os.path.basename(command[0]).lower():
+            command.append('-f')
 
         for item in items:
             if item['initial_key'] and not overwrite:
                 continue
 
             try:
-                output = util.command_output([bin, '-f', item.path])
+                output = util.command_output(command + [util.syspath(
+                                                        item.path)]).stdout
             except (subprocess.CalledProcessError, OSError) as exc:
-                self._log.error(u'execution failed: {0}', exc)
+                self._log.error('execution failed: {0}', exc)
+                continue
+            except UnicodeEncodeError:
+                # Workaround for Python 2 Windows bug.
+                # https://bugs.python.org/issue1759845
+                self._log.error('execution failed for Unicode path: {0!r}',
+                                item.path)
                 continue
 
-            key_raw = output.rsplit(None, 1)[-1]
             try:
-                key = key_raw.decode('utf8')
+                key_raw = output.rsplit(None, 1)[-1]
+            except IndexError:
+                # Sometimes keyfinder-cli returns 0 but with no key, usually
+                # when the file is silent or corrupt, so we log and skip.
+                self._log.error('no key returned for path: {0}', item.path)
+                continue
+
+            try:
+                key = util.text_string(key_raw)
             except UnicodeDecodeError:
-                self._log.error(u'output is invalid UTF-8')
+                self._log.error('output is invalid UTF-8')
                 continue
 
             item['initial_key'] = key
-            self._log.info(u'added computed initial key {0} for {1}',
+            self._log.info('added computed initial key {0} for {1}',
                            key, util.displayable_path(item.path))
 
             if write:
